/*
 * GameList Class 
 */

#include "bcmserv.h"


GameList::GameList ()
{
	stat_maxplayers = 0;
	stat_maxplayerhost[0] = 0;
	stat_maxplayerport[0] = 0;
	stat_maxplayername[0] = 0;
	stat_countedgamesv4 = 0;
	stat_countedgamesv6 = 0;
};


GameList::~GameList ()
{
};


void
GameList::Update (int nr, struct game_entry *ge)
{
    struct game_entry *georg = 0;

    georg = (struct game_entry *) List::Get (nr);
    if (georg == 0) {
        delete ge;
        return;
    }
    *georg = *ge;
	
	if (ge->curplayers > stat_maxplayers) { /* this holds the high score */
		stat_maxplayers = ge->curplayers;
		strncpy (stat_maxplayerhost, ge->host, 128);
		strncpy (stat_maxplayerport, ge->port, 10);
		strncpy (stat_maxplayername, ge->name, 64);
	}
	
    delete ge;
};


void
GameList::Add (struct _sockaddr *Addr, char *data)
{
    int i,
        i1,
        i2,
        i3;
    char *pos;
    struct game_entry *ge = new struct game_entry;

    for (i = 0; i < LEN_SERVERNAME; i++) {
        ge->host[i] = 0;
        if (i < LEN_PORT)
            ge->port[i] = 0;
        if (i < LEN_GAMENAME)
            ge->name[i] = 0;
    };

    ge->maxplayers = 0;
    ge->curplayers = 0;

    ge->timestamp = time (NULL);
    dns_filladdr (ge->host, LEN_SERVERNAME, ge->port, LEN_PORT, ((struct sockaddr*)Addr)->sa_family, Addr);

    for (pos = data; pos != 0;) {
        // go throught the incoming data 
        switch (pos[0]) {
        case ('P'):
            sscanf (pos + 1, "%d,%d", &i1, &i2);
            ge->curplayers = i1;
            ge->maxplayers = i2;
            break;
        case ('V'):
            sscanf (pos + 1, "%d.%d.%d", &i1, &i2, &i3);
            ge->version.major = i1;
            ge->version.minor = i2;
            ge->version.sub = i3;
            break;
        case ('S'):
            sscanf (pos + 1, "%d,%d,%d", &i1, &i2, &i3);
            ge->state = i1;
            ge->multitype = i2;
            ge->gametype = i3;
            break;
        case ('I'):
            if (pos[1] == '4')
                ge->ai_family = PF_INET;
            else if (pos[1] == '6')
                ge->ai_family = PF_INET6;
            break;
        case ('N'):
            for (i = 1; (i - 1 < LEN_GAMENAME) && (pos[i] != 0)
                 && (pos[i] != '\n'); i++)
                ge->name[i - 1] = pos[i];
            ge->name[i - 1] = 0;
            break;
        case ('H'):
            for (i = 1; (i - 1 < LEN_SERVERNAME) && (pos[i] != 0)
                 && (pos[i] != '\n'); i++)
                ge->host[i - 1] = pos[i];
            ge->host[i - 1] = 0;
            break;
        case ('O'):
            for (i = 1; (i - 1 < LEN_PORT) && (pos[i] != 0)
                 && (pos[i] != '\n'); i++)
                ge->port[i - 1] = pos[i];
            ge->port[i - 1] = 0;
            break;
        }
        pos = strchr (pos, '\n');
        if (pos != 0)
            pos++;
        if (pos != 0 && pos[0] == 0)
            pos = 0;
    }
	
    if (ge->host[0] == 0 || ge->port[0] == 0 || ge->maxplayers <= 0) {
        printf ("something wrong\n");
        delete ge;
        return;
    }

    printf ("Add oky\n");
    Add (ge);
};


void
GameList::Add (struct game_entry *ge)
{
    int nr;
    nr = Search (ge->host, ge->port);
	
    if (nr != -1) 
        Update (nr, ge);
    else {
		if (ge->ai_family == PF_INET)
			stat_countedgamesv4++;
		else
			stat_countedgamesv6++;
		
        List::Add (ge);
	}
};


void
GameList::Del (char *host, char *port)
{
    int nr;

    nr = Search (host, port);
    if (nr != -1)
        Del (nr);
};


void
GameList::Del (int nr)
{
    struct game_entry *ge = 0;

    ge = Get (nr);
    delete ge;
    List::Del (nr);
};


void
GameList::PrintList ()
{
    int i;

    cout << "****** GameList\n";

    for (i = 0; i < count; i++) {
        struct game_entry *ge = Get (i);
        cout << "Host : " << ge->host << " Port : " << ge->port << " GameName : " << ge->name << "\n";
    }

};


void
GameList::Del (struct _sockaddr *Addr)
{
    char host[LEN_SERVERNAME], port[LEN_PORT];

    host[0] = 0;
    port[0] = 0;

    dns_filladdr (host, LEN_SERVERNAME, port, LEN_PORT, PF_UNSPEC, Addr);

    Del (host, port);
};


struct game_entry *
GameList::Get (int nr)
{
    return (struct game_entry *) List::Get (nr);
};


int
GameList::Search (char *host, char *port)
{
    struct game_entry *ge;
    int i,
        y;

    for (i = 0, y = -1; (i < count) && (y == -1); i++) {
        ge = (struct game_entry *) List::Get (i);
        if (strcmp (ge->host, host) == 0 && strcmp (ge->port, port) == 0)
            y = i;
    }

    return y;
};


void
GameList::CleanUp ()
{
    int i,
        y;
    struct game_entry *ge;
    time_t timestamp;

    timestamp = time (NULL);

    for (i = 0, y = -1; (i < count) && (y == -1); i++) {
        ge = (struct game_entry *) List::Get (i);
        if (timestamp - ge->timestamp >= DEFAULT_GAMETIMEOUT)
            y = i;
    }

    if (y != -1)
        Del (y);
};


void
GameList::CreatePacket (int nr, char *data, int len)
{
    char tmp[1024];
    struct game_entry *ge;

    data[0] = 0;

    if (nr >= count)
        return;
    ge = (struct game_entry *) Get (nr);

    sprintf (tmp, "P%d,%d\n", ge->curplayers, ge->maxplayers);
    sprintf (tmp, "%sV%d.%d.%d\n", tmp, ge->version.major, ge->version.minor, ge->version.sub);
    sprintf (tmp, "%sS%d,%d,%d\n", tmp, ge->state, ge->multitype, ge->gametype);
    if (ge->ai_family == PF_INET)
        sprintf (tmp, "%sI4\n", tmp);
    else
        sprintf (tmp, "%sI6\n", tmp);
    sprintf (tmp, "%sN%s\n", tmp, ge->name);
    sprintf (tmp, "%sH%s\n", tmp, ge->host);
    sprintf (tmp, "%sO%s\n", tmp, ge->port);

    strncpy (data, tmp, len);
};
