/* tcp.c Version 0.0.1
 * TCP Network Code
 */

#include "tcpip.h"


void
unblock_socket (int sock)
{

    int flags;

#ifdef _WIN32
/* Windows Code of the tcp part 
   is not needed, we will work over select */

#else
/* Linux Code of the tcp part */

    /* get current socket flags */
    if ((flags = fcntl (sock, F_GETFL)) == -1) {
        perror ("GETFL");
    }

    /* set socket to non-blocking */
    flags |= O_NDELAY;
    if (fcntl (sock, F_SETFL, flags) == -1) {
        perror ("SETFL");
    }
#endif
};


/* close TCP Socket */
void
tcp_close (int sock)
{
    close (sock);
};


/* send date to an TCP socket*/
int
tcp_send (int sock, char *data, int len)
{
    return send (sock, data, len, 0);
};


/* open the server */
int
tcp_server (char *port)
{
    int err,
      i,
      sock = -1;

#ifdef _WIN32
/* Windows Code of the tcp part */

    struct sockaddr_in adresse;
    size_t sockaddr_insize = sizeof (struct sockaddr_in);
    struct servent *sent;

    printf ("TCP_Server (%s)\n", port);

    /* Socket an Filediscriptor anbinden */
    if ((sock = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
        fprintf (stderr, "Error: Could not get socket.\n");
        return -1;
    }

    sent = getservbyname (port, "tcp");
    if (sent == NULL) {
        adresse.sin_port = htons (atoi (port));
    }
    else
        adresse.sin_port = sent->s_port;
    adresse.sin_family = AF_INET;
    memset (&adresse.sin_addr, 0, sizeof (adresse.sin_addr));

    if (bind (sock, (struct sockaddr *) &adresse, sizeof (adresse))) {
        fprintf (stderr, "Error: Could not bind Socket.\n");
        return -1;
    }

    printf ("  got bind\n");

    if (listen (sock, 1)) {
        fprintf (stderr, "Error: Could not go on listen.\n");
        return -1;
    }

    printf ("  got listen\n");


#else
/* Linux Code of the tcp part */

    struct addrinfo hints,
     *res,
     *sres;

    bzero (&hints, sizeof (struct addrinfo));

    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_STREAM;

    if ((err = getaddrinfo (NULL, port, &hints, &res)) == 0) {

        sres = res;

        if ((sock = socket (sres->ai_family, sres->ai_socktype, 0)) < 0) {
            printf ("%d=socket(PF_INET6, SOCK_DGRAM, 0)\n", sock);
            sock = -1;
            freeaddrinfo (res);
            return -1;
        }

        i = 1;

        setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (i));

        if ((err = bind (sock, sres->ai_addr, sres->ai_addrlen)) < 0) {
            close (sock);
            printf ("%d = bind (sd, res->ai_addr, res->ai_addrlen)\n", err);
            sock = -1;
            freeaddrinfo (res);
            return -1;
        }

        listen (sock, 5);

        unblock_socket (sock);

        freeaddrinfo (res);
    }
    else {
        printf ("getaddrinfo:%s\n", gai_strerror (err));
    }

    printf ("Listensocket : %d\n", sock);
#endif

    return sock;
};


/* connect to an server */
int
tcp_connect (char *host, char *port)
{
    int err,
      sock = -1;

#ifdef _WIN32
/* Windows Code of the tcp part */
    struct hostent *he;
    struct in_addr ip_adr;
    struct sockaddr_in adresse;

    printf ("TCP_Connect (%s, %s)\n", host, port);

    he = gethostbyname (host);

    if (he == NULL) {
        fprintf (stderr, "Error: Could not get hostname (%s)\n", host);
        return -1;
    }

    printf ("TCP_Connect to %s:%s\n", host, port);

    /* adressen vorbereiten */
    adresse.sin_family = AF_INET;
    adresse.sin_port = htons (atoi (port));

    memcpy (&adresse.sin_addr, he->h_addr_list[0], sizeof (adresse.sin_addr));

    /* Socket an Filediscriptor binden, aufbaun der verbindung */
    if ((sock = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
        fprintf (stderr, "ERROR: could not get socket.\n");
        return -1;
    }

    if (connect (sock, (struct sockaddr *) &adresse, sizeof (adresse))) {
        fprintf (stderr, "ERROR: could not connect.\n");
        return -1;
    }

#else
/* Linux Code of the tcp part */

    struct addrinfo hints,
     *res,
     *sres;

    printf ("TCP_Connect to %s:%s\n", host, port);

    bzero (&hints, sizeof (struct addrinfo));

    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_STREAM;

    if ((err = getaddrinfo (host, port, &hints, &res)) == 0) {

        sres = res;

        if ((sock = socket (sres->ai_family, sres->ai_socktype, 0)) < 0) {
            printf ("%d=socket(PF_INET, SOCK_DGRAM, 0)\n", sock);
            sock = -1;
            freeaddrinfo (res);
            return -1;
        }

        if (connect (sock, res->ai_addr, res->ai_addrlen) < 0) {
            printf ("connect error\n");
            close (sock);
            freeaddrinfo (res);
            return -1;
        }

        unblock_socket (sock);

        freeaddrinfo (res);
    }
    else {
        printf ("getaddrinfo:%s\n", gai_strerror (err));
    }
#endif

    return sock;
};


/* accept connections if there are any */
int
tcp_accept (int tcpsocket, char *addr, int len)
{
    int sock;

#ifdef _WIN32
/* Windows Code of the tcp part */
    fd_set listenset;
    struct sockaddr_in remoteaddr;
    int addrlen;
    int i;
    struct timeval tval;


    FD_ZERO (&listenset);
    FD_SET (tcpsocket, &listenset);

    tval.tv_sec = 0;
    tval.tv_usec = 100;
    if (select (tcpsocket + 1, &listenset, NULL, NULL, &tval)) {
        // got a connection
        printf ("TCP_Accept   03\n");

        addrlen = sizeof (remoteaddr);

        if ((sock = accept (tcpsocket,
                            (struct sockaddr *) &remoteaddr, &addrlen)) == -1) {
            perror ("accept");
            sock = -1;
        }
        else {
            strncpy (addr, inet_ntoa (remoteaddr.sin_addr), len);
            printf ("selectserver: new connection from %s\n", addr);
        }
    }
    else
        sock = -1;

#else

/* Linux Code of the tcp part */
    struct sockaddr_in cAddr;
    size_t addrlen = sizeof (struct sockaddr_in);

    sock = accept (tcpsocket, (struct sockaddr *) &cAddr, &addrlen);

    if (sock > 0) {
        unblock_socket (sock);

        strncpy (addr, inet_ntoa (((struct sockaddr_in *) &cAddr)->sin_addr),
                 len);
    }
    if (sock < 0 && errno != EBADF && errno != EWOULDBLOCK) {
        perror ("tcp_accept");
        printf ("  %d\n", sock);
    }
#endif

    return sock;
};


/* gets some data from socket, if there are any data aviable */
int
tcp_get (int tcpsocket, int len, char *data)
{
    int len_ = len;

#ifdef _WIN32
/* Windows Code of the tcp part */

    fd_set rfds;
    struct timeval tval;
    int i;

    FD_ZERO (&rfds);
    FD_SET (tcpsocket, &rfds);

    tval.tv_sec = 0;
    tval.tv_usec = 100;

    if (select (tcpsocket + 1, &rfds, NULL, NULL, &tval)) {
        len_ = recv (tcpsocket, data, len, 0);

        if (i <= 0)
            return -1;
        data[len_] = 0;
    }
    else
        return 0;

#else
/* Linux Code of the tcp part */
    len_ = read (tcpsocket, data, len_);

    if (len_ < 0 && errno == EAGAIN)
        len_ = 0;

    else if (len_ < 0 && errno != EAGAIN) {
        perror ("tcp_get");
        len_ = -1;
    }

    else if (len_ == 0)
        len_ = -1;
#endif

    return len_;
}
