/* udp.c Version 0.0.2
 * Server and Client Code for UDP Connection's
 */

#include "tcpip.h"

extern char *dns_net_getip (char *host);
extern int dns_filladdr (char *host, int hostlen, char *port, int portlen, int ai_family,
                         struct _sockaddr *sAddr);

extern int udp_get (int sock, char *text, int len, struct _sockaddr *sAddr, int ai_family);
extern int udp_server (char *port, int ai_family);
extern void udp_send (int sock, char *text, int len, struct _sockaddr *sAddr, int ai_family);
extern void udp_close (int sock);


/* closes an existing udp server */
void
udp_close (int sock)
{
    if (sock != -1)
        close (sock);
    sock = -1;
};




/* send text to someone */
void
udp_send (int sock, char *text, int len, struct _sockaddr *sAddr, int ai_family)
{
    int addrlen;

    if (ai_family == PF_INET)
        addrlen = sizeof (struct sockaddr_in);
#ifndef _WIN32
    else
        addrlen = sizeof (struct sockaddr_in6);
#endif

    if (sendto (sock, text, len, 0, (struct sockaddr *) sAddr, addrlen) == -1)
        perror ("udp_send :");

};


int
udp_server (char *port, int ai_family)
{
#ifdef _WIN32

    int sock;
    struct sockaddr_in sAddr;   // my address information

    if ((sock = socket (ai_family, SOCK_DGRAM, 0)) == -1) {
        perror ("udp_server: socket");
        return -1;
    }

    sAddr.sin_family = AF_INET; // host byte order
    sAddr.sin_port = htons (atoi (port)); // short, network byte order
    sAddr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP

    memset (&(sAddr.sin_zero), '\0', 8); // zero the rest of the struct

    if (bind (sock, (struct sockaddr *) &sAddr, sizeof (struct sockaddr)) == -1) {
        perror ("udp_server: bind");
        return -1;
    }

#else
    struct addrinfo hints,
     *res,
     *sres;
    int err,
      sock,
      ai_family_;

    bzero (&hints, sizeof (struct addrinfo));

    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = ai_family;
    hints.ai_socktype = SOCK_DGRAM;

    ai_family_ = 0;

    if ((err = getaddrinfo (NULL, port, &hints, &res)) == 0) {
        sres = res;
        while ((ai_family_ == 0) && (sres)) {
            if (sres->ai_family == ai_family || ai_family == PF_UNSPEC)
                ai_family_ = sres->ai_family;
            else
                sres = sres->ai_next;
        }

        if (sres == NULL)
            sres = res;

        ai_family_ = sres->ai_family;
        if (ai_family_ != ai_family && ai_family != PF_UNSPEC) {
            // ai_family is not identic
            freeaddrinfo (res);
            return -1;
        }

        if ((sock = socket (sres->ai_family, SOCK_DGRAM, 0)) < 0) {
            perror ("UDP_Server (socket):");
            freeaddrinfo (res);
            return -1;
        }

        if ((err = bind (sock, sres->ai_addr, sres->ai_addrlen)) < 0) {
            perror ("UDP_Server (bind):");
            close (sock);
            freeaddrinfo (res);
            return -1;
        }

        freeaddrinfo (res);
    }
    else {
        sock = -1;
        printf ("UDP_Server (getaddrinfo):%s\n", gai_strerror (err));
    }
#endif

    return sock;
};

/*
   gets some text
   RESULT: 0 for nothing on there 
*/
int
udp_get (int sock, char *text, int len, struct _sockaddr *sAddr, int ai_family)
{
    int clen,
      msglen;
    fd_set sockset;
    struct timeval tval;

    /* what version of tcp/ip we're using */
    if (ai_family == AF_INET)
        clen = sizeof (struct sockaddr_in);
#ifndef _WIN32
    else
        clen = sizeof (struct sockaddr_in6);
#endif

    bzero (text, len);

    // check if we have got any data
    FD_ZERO (&sockset);
    FD_SET (sock, &sockset);

    tval.tv_sec = 0;
    tval.tv_usec = 100;

    msglen = 0;

    if (select (sock + 1, &sockset, NULL, NULL, &tval)) {

        msglen = recvfrom (sock, text, len, 0, (struct sockaddr *) sAddr, &clen);
        if (msglen < 0)
            return 0;

        if ((msglen >= 0) && (msglen < len))
            text[msglen] = 0;
    }
    return msglen;
};
